#include "OpenSPC.h"
#include "config.h"
#ifdef HAVE_ESD
#include <esd.h>
#endif
#ifdef HAVE_LINUX_SOUND
#include <sys/soundcard.h>
#endif
#include "tty.h"

struct riffhdr
{
  char    riff[4];      /* 'RIFF' */
  long    filesize;     /* minus these 8 bytes */
  char    rifftype[4];     /* 'WAVE' */
};

struct chunk
{
  char    chunkid[4];   /* 4-character chunk ID */
  long    chunksize;    /* minus these 8 bytes */
};

struct fmt
{
  short   fmttag;       /* 1= WAVE_FORMAT_PCM */
  short   channels;     /* 1= mono, 2= stereo */
  long    samplerate;   /* 11025, 22050 ... */
  long    bytespersec;  /* bytes per second */
  short   blockalign;   /* datasize*channels/8 */
  short   datasize;     /* 8= 8-bit, 16= 16-bit */
};

union pcm
{
  unsigned char *d8;
  signed short  *d16;
};

void Shutdown(void);

void Usage(void)
{
  fprintf(stderr,"Usage: OpenSPClite [options] <filename>\n");
  fprintf(stderr,"Where <filename> is any .SPC, .ZST, .SP#, or .ZS# file\n\n");
  fprintf(stderr,"Available switches:\n");
  fprintf(stderr,"    -i		Enables IT and disables WAV dumping.  	[WAV default]\n");
  fprintf(stderr,"    -u xxx	Specify update rate			[100 default]\n");
  fprintf(stderr,"    -t x:xx	Specify a time limit			[unlimited default]\n");
  fprintf(stderr,"    -n x:xx	Specify a null-time limit		[0:05 default]\n");
  fprintf(stderr,"    -d xxxxxxxx Voices to disable (1-8)			[none default]\n");
  fprintf(stderr,"    -f xxxxx	Specify output frequency (WAV only)	[44100 default]\n");
  fprintf(stderr,"    -8		Enable 8 bit output (WAV only)		[16 bit default]\n");
  fprintf(stderr,"    -m		Specify mono output (WAV only)		[stereo default]\n");
  fprintf(stderr,"    -r xxx	Specify rows per pattern (IT only)	[200 default]\n");
  fprintf(stderr,"    -w		generate a .WAV file\n");
#ifdef HAVE_ESD
  fprintf(stderr,"    -e		esdout\n");
  fprintf(stderr,"    -h xxx	Specify hostname for esd\n");
#endif
#ifdef HAVE_LINUX_SOUND
  fprintf(stderr,"    -l		linuxout\n");
#endif
  fprintf(stderr,"    -p		pipe output to stdout\n");
  fprintf(stderr,"    -P		pipe input from stdin\n");
  fprintf(stderr,"    -a		disable rate 'tweak' and use specified.\n");
}

FILE *wavtmp;
int wavcnt, freq, stereo, bit8;
int wavout=0, pipein=0, pipeout=0;
#ifdef HAVE_ESD
int esdout=0;
int sock = -1;
esd_format_t format = 0;
char *host = NULL;
#endif
#ifdef HAVE_LINUX_SOUND
int linuxout=0, linuxadjust=1;
#endif
char fn2[64];

int WAVStart(void)
{
  struct riffhdr hdr={"RIFF",0,"WAVE"};
  struct chunk chinfo;
  struct fmt fmt;
  int size;
  
  wavtmp=fopen("wavtemp.tmp","wb+");
  if(wavtmp==NULL) return 1;

  wavcnt=0; size=1;

  fmt.fmttag=1;
  fmt.channels=1;
  fmt.datasize=8;

  if(stereo)
  {
    fmt.channels=2;
    size<<=1;
  }
  if(bit8==0)
  {
    fmt.datasize=16;
    size<<=1;
  }

  hdr.filesize=-1; // fill in later

  fwrite(&hdr,sizeof(struct riffhdr),1,wavtmp);
  strcpy(chinfo.chunkid,"fmt ");
  chinfo.chunksize=sizeof(struct fmt);
  fwrite(&chinfo,sizeof(struct chunk),1,wavtmp);

  fmt.samplerate=SNDfreq;
  fmt.bytespersec=SNDfreq*size;
  fmt.blockalign=size;
  fwrite(&fmt,sizeof(struct fmt),1,wavtmp);
  strcpy(chinfo.chunkid,"data");
  chinfo.chunksize=-1; // fill in later
  fwrite(&chinfo,sizeof(struct chunk),1,wavtmp);

  return 0;
}

#ifdef HAVE_LINUX_SOUND
int OpenLinuxOut(void)
{
  int format, tmp;

  linuxout=open("/dev/dsp",O_WRONLY,0);
  if (linuxout==0)
  {
    fprintf(stderr, "/dev/dsp didnt open.\n");
    return 1;
  }
  format=(bit8)?AFMT_U8:AFMT_S16_LE;

// i'm not sure how to check if these ioctls FAIL quite yet. add later.
// if /dev/dsp opens there really should be no serious failures, however
// the user may get an unexpected result if their card doesnt support some
// option combo...

  ioctl(linuxout,SNDCTL_DSP_SETFMT,&format);
  ioctl(linuxout,SNDCTL_DSP_STEREO,&stereo);
  tmp=freq;
  ioctl(linuxout,SNDCTL_DSP_SPEED,&tmp);

  fprintf(stderr,"Actual rate soundcard reported: %d\n", tmp);

  if (tmp==freq) fprintf(stderr, "This matches what I requested. Joy!\n");
  else
  {
    if (linuxadjust)
    {
      fprintf(stderr, "Frequency adjusted to match.\n");
      freq=tmp;
    }
    else
    {
      fprintf(stderr, "Ignoring soundcard's status report and using original specified frequency.\n");
    }
  }
  return 0;
}
#endif

void WAVUpdate(void)
{
  int bytes=1;
  
  if(bit8==0) bytes<<=1;
  if(stereo) bytes<<=1;

  if (pipeout)
    fwrite(SNDoptbuf,bytes,SNDmixlen,stdout);
  if (wavout)
  {
    fwrite(SNDoptbuf,bytes,SNDmixlen,wavtmp);
    wavcnt++;
  }
#ifdef HAVE_ESD
  if (esdout)
    write(sock,SNDoptbuf,bytes*SNDmixlen);
#endif
#ifdef HAVE_LINUX_SOUND
  if (linuxout)
    write(linuxout,SNDoptbuf,bytes*SNDmixlen);
#endif
}

void WAVWrite(char *fn)
{
  int size=1;

  wavcnt*=SNDmixlen;
  if(bit8==0) size<<=1;
  if(stereo)  size<<=1;

  wavcnt*=size;
// reuse 'size' variable since we're done with it:
  size=(4+sizeof(struct chunk)*2+sizeof(struct fmt)+wavcnt);

  fseek(wavtmp, 4, SEEK_SET);
  fwrite(&size, 4, 1, wavtmp);

  fseek(wavtmp, 40, SEEK_SET);
  fwrite(&wavcnt, 4, 1, wavtmp);

  fclose(wavtmp);
  rename("wavtemp.tmp", fn);
}

int main(int argc, char **argv)
{
  int urate; // int cursx,cursy;
  int seconds,limit,nlimit,ntime,done,voices,v;
  int i;
  fprintf(stderr,"\nOpenSPClite-%s Linux (based on Lite/Direct To Disk version)\n", VERSION);
  fprintf(stderr," Linux port and general code tweaking by Chris Timmerberg (ctimmerb@usa.net)\n");
  fprintf(stderr," Additional modifications by Nicolas NOBLE aka Pixel (pixels@multimania.com)\n");
  fprintf(stderr," Original Author: Butcha\n");
  fprintf(stderr," Other contributors: Cyber Warrior X, Crono, Kadar\n");
  fprintf(stderr," Thanks to: Savoury SnaX and Charles Bilyu for SNEeSe's SPC core\n\n");
  fn2[0]=0; SNDmix=0;
  urate=100;	// Default 100Hz update rate
  freq=44100;	// Default 44100Hz freq (WAV only)
  stereo=1;     // Default stereo
  bit8=0;               // Default 8 bit
  ITrows=200;	// Default 200 rows/pattern (IT only)
  limit=0;	// Default no time limit
  nlimit=5;	// Default 5s null-time limit
  voices=0;	// Default no voices disabled
  for(i=1;i<argc;i++)
  {
    if(argv[i][0]=='-')
      switch(argv[i][1])
      {
      case 'i': ITdump=1; break;
      case 'u': i++; urate=atoi(argv[i]);  break;
      case 'f':	i++; freq=atoi(argv[i]);   break;
      case '8':	bit8=1;   break;
      case 'm':	stereo=0; break;
      case 'r':	i++; ITrows=atoi(argv[i]); break;
      case 't':	i++;
        sscanf(argv[i],"%d:%d",&limit,&seconds);
        limit*=60;
        limit+=seconds;
        break;
      case 'n':	i++;
        sscanf(argv[i],"%d:%d",&nlimit,&seconds);
        nlimit*=60;
        nlimit+=seconds;
        break;
      case 'd':	i++;
        for(v=0;argv[i][v]!=0;v++)
          voices|=1<<(argv[i][v]-'1');
        break;
      case 'w': wavout=1;   SNDmix=1;  break;
#ifdef HAVE_ESD
      case 'e': esdout=1; SNDmix=1; break;
      case 'h': i++; host=argv[i]; break;
#endif      
#ifdef HAVE_LINUX_SOUND
      case 'l': linuxout=1; SNDmix=1;  break;
      case 'a': linuxadjust=0;      break;
#endif
      case 'p': pipeout=1;  SNDmix=1;  break;
      case 'P': pipein=1; fprintf(stderr, "Pipe input not done yet!\n"); break;
      default:	fprintf(stderr,"Warning, unrecognized option '-%c'!\n",argv[i][1]);
      }
    else strcpy(fn2,argv[i]);
  }
  if(fn2[0]==0)
  {
    Usage();
    exit(0);
  }

// if no output options are specified, turn on esdout, then linuxout, and waveout
// otherwise
  if (pipeout==0 && 
#ifdef HAVE_ESD
    esdout==0 &&
#endif
#ifdef HAVE_LINUX_SOUND
    linuxout==0 && 
#endif
    ITdump==0) {
#ifdef HAVE_ESD
    esdout=1;
#else
#ifdef HAVE_LINUX_SOUND
    linuxout=1;
#else
    wavout=1;
#endif
#endif
    SNDmix=1; }
  tty_init();

  if(ITdump)
  {
    if(urate>102)
    {
      fprintf(stderr,"Error: Can only record to IT at 102 updates/sec or less!\n");
      exit(1);
    }
  }
  else	if(urate>1000)
  {
    fprintf(stderr,"Error: Max update rate is 1000\n");
    exit(1);
  }

  fprintf(stderr,"\nFilename:	%s\n",fn2);
  fprintf(stderr,"Time limit:	");
  if(limit)
    fprintf(stderr,"%d:%02d\n",limit/60,limit%60);
  else 	fprintf(stderr,"none\n");
  fprintf(stderr,"Null-Time limit:	");
  if(nlimit)
    fprintf(stderr,"%d:%02d\n",nlimit/60,nlimit%60);
  else 	fprintf(stderr,"none\n");
  fprintf(stderr,"Voices disabled: ");
  if(voices)
  {
    for(v=0;v<8;v++)
      if(voices&(1<<v))
        fprintf(stderr,"%d",v+1);
    fprintf(stderr,"\n");
  }
  else fprintf(stderr,"none\n");
  fprintf(stderr,"Output format(s):	");
  if (ITdump) fprintf(stderr," IT");
  if (wavout) fprintf(stderr," WAV");
  if (pipeout) fprintf(stderr," PIPE");
#ifdef HAVE_ESD
  if (esdout) fprintf(stderr, "ESounD-Out");
#endif
#ifdef HAVE_LINUX_SOUND
  if (linuxout) fprintf(stderr," LinuxOut");
#endif
  fprintf(stderr, "\n");
  if(ITdump)
  {
    fprintf(stderr,"IT Parameters:\n");
    fprintf(stderr,"	Update rate:	%d\n",urate);
    fprintf(stderr,"	Rows/pattern:	%d\n",ITrows);
  }
  if(SNDmix)
  {
    fprintf(stderr,"WAV/PIPE/LinuxOut/ESounD-Out Parameters:\n");
    fprintf(stderr,"	Update rate:	%d\n",urate);
    fprintf(stderr,"	Frequency:	%d\n",freq);
    fprintf(stderr,"	Quality:	");
    if(bit8)
      fprintf(stderr,"8 bit ");
    else    fprintf(stderr,"16 bit ");
    if(stereo)
      fprintf(stderr,"stereo\n");
    else 	fprintf(stderr,"mono\n");
  }

#ifdef HAVE_ESD
  if (esdout) {
    format = ((bit8 ? ESD_BITS8 : ESD_BITS16) |
              (stereo ? ESD_STEREO : ESD_MONO) | ESD_STREAM | ESD_PLAY);
//    printf("opening socket, format = 0x%08x at %d Hz\n", format, freq);
    sock = esd_play_stream_fallback(format, freq, host, NULL);
    if (sock <= 0) {
      fprintf(stderr, "ESounD-Out did not initialize properly. Operating in linuxout mode.\n");
      esdout = 0;
      linuxout = 1;
    }
  }
#endif

#ifdef HAVE_LINUX_SOUND
// linuxout potentially changes the freqency: this open must occur before the
// spc/snd stuff.

  if (linuxout&&OpenLinuxOut())
  {
    fprintf(stderr,"LinuxOut did not initialize properly. Operating in disk-only mode.\n");
    linuxout = 0;
    wavout = 1;
  }
#endif

  if(SPCInit(fn2))	//Reset SPC and load state
  {
    fprintf(stderr,"Failed to initialize emulation!\n");
    exit(1);
  }
  if(SNDInit(freq,urate,1,stereo,bit8))
  {
    fprintf(stderr,"Init Sound failed!\n");
    exit(1);
  }
  SNDvmask=~voices;
  if(ITdump&&ITStart())
  {
    fprintf(stderr,"Failed to init IT pattern buffers! Disabling IT output.\n");
    ITdump=0;
  }
  if(wavout&&WAVStart())
  {
    fprintf(stderr,"Failed to init WAV dump! Disabling WAV output.\n");
    wavout=0;
  }
  if (pipeout==0 && 
#ifdef HAVE_ESD
      esdout==0 &&
#endif
#ifdef HAVE_LINUX_SOUND
      linuxout==0 &&
#endif
      wavout==0 && ITdump==0)
  {
    fprintf(stderr,"No output methods remain enabled. Exiting.\n");
    exit(1);
  }
        if(SPCname[0]||SPCtname[0]||SPCdname[0]||SPCcomment[0]||SPCdate[0])
        {
                printf("ID info:\n");
                printf("        Song name:              %s\n",SPCname);
                printf("        Game name:              %s\n",SPCtname);
                printf("        Person who Dumped SPC:  %s\n",SPCdname);
                printf("        Comments:               %s\n",SPCcomment);
                printf("        Dumped on date:         %s\n",SPCdate);
        }

  SNDNoteOn(SPC_DSP_Buffer[0x4c]);
  if(limit)
    fprintf(stderr,"Progress: ");
  else	fprintf(stderr,"Play time: ");
  //        fflush(stdout);
  //        cursx=wherex();
  fprintf(stderr,"\nPress spacebar to abort playback\n");
  //        cursy=wherey()-2;
  seconds=SNDratecnt=done=0;
  SNDNoteOn(SPC_DSP_Buffer[0x4c]); //Turn on voices that were on when saved
  SNDcrc = 0;
  ntime = 0;
  while(!done)
  {
    SNDMix();
    if(ITdump)
    {
      if(ITUpdate())
        break;
    }
    if(SNDmix)
      WAVUpdate();

    if(SNDratecnt>=urate)
    {
      if(tty_getchar()==32) done=1;
      SNDratecnt-=urate;
      seconds++;
      if (SNDcrc == 0) {
        ntime++;
      } else {
        ntime = 0;
      }
      //	        	gotoxy(cursx,cursy);
      if (nlimit && (ntime == nlimit)) done = 1;
      if(limit)
      {
        if(seconds==limit)
          done=1;
        fprintf(stderr,"%3d%%\r",seconds*100/limit);
      }
      else
        fprintf(stderr,"%2d:%02d\r",seconds/60,seconds%60);
      //	                fflush(stdout);
      SNDcrc = 0;
    }
  }
  Shutdown();
  return(0);
}

void Shutdown(void)
{
  int i=0;

  if(ITdump || wavout)
  {
    fprintf(stderr,"\n\nSaving file...\n");
    for(i=0;i<64;i++)
      if(fn2[i]==0)
        break;
    for(;i>0;i--)
      if(fn2[i]=='.') break;
  }
  if(ITdump)
  {
    strcpy(&fn2[i+1],"IT");
    if(ITWrite(fn2))
      fprintf(stderr,"Failed at writing %s.\n",fn2);
    else
      fprintf(stderr,"Wrote %s successfully.\n",fn2);
  }
  if(wavout)
  {
    strcpy(&fn2[i+1],"WAV");
    WAVWrite(fn2);
    fprintf(stderr,"Wrote %s successfully.\n",fn2);
  }
#ifdef HAVE_ESD
  if(esdout) close(sock);
#endif
#ifdef HAVE_LINUX_SOUND
  if(linuxout) close(linuxout);
#endif
  fprintf(stderr,"Done!\n"); tty_done();
}
